// PG Tools, (c)2021-2022 Tomas Fabian
// 
// May only be used for non-commercial educational purposes at VSB-TUO, FEECS,
// Dept. of Computer Science and may not be shared with any third parties in any form.

#ifndef VECTOR2_H_
#define VECTOR2_H_

#include <array>

#include "vector3.h"

/*! \struct Vector2
\brief Dvourozmrn (2D) vektor.

Implementace dvouslokovho relnho vektoru podporujc zkladn matematick operace.

\note
Vektor se povauje za sloupcov, pestoe je v komentch pro jednoduchost uvdn jako dkov.

\code{.cpp}
Vector2 v = Vector2( 2.0f, -7.8f );
v.Normalize();
\endcode

\author Tom Fabin
\version 0.95
\date 2007-2021
*/
class Vector2
{
public:
	union	// anonymn unie
	{
		struct
		{
			float x; /*!< Prvn sloka vektoru. */
			float y; /*!< Druh sloka vektoru. */			
		};

		struct
		{
			float u; /*!< Prvn sloka vektoru. */
			float v; /*!< Druh sloka vektoru. */
		};

		std::array<float, 2> data; /*!< Pole sloek vektoru. */		
	};

	//! Vchoz konstruktor.
	/*!
	Inicializuje vechny sloky vektoru na hodnotu nula,
	\f$\mathbf{v}=\mathbf{0}\f$.
	*/
	Vector2() : x( 0 ), y( 0 ) { }	

	//! Obecn konstruktor.
	/*!
	Inicializuje sloky vektoru podle zadanch hodnot parametr,
	\f$\mathbf{v}=(x,y,z)\f$.

	\param x prvn sloka vektoru.
	\param y druh sloka vektoru.	
	*/
	Vector2( const float x, const float y ) : x( x ), y( y ) { }

	//! Konstruktor z pole.
	/*!
	Inicializuje sloky vektoru podle zadanch hodnot pole,

	\param v ukazatel na prvn sloka vektoru.	
	*/
	Vector2( const float * v );

	//! L2-norma vektoru.
	/*!
	\return x Hodnotu \f$\mathbf{||v||}=\sqrt{x^2+y^2+z^2}\f$.
	*/
	float L2Norm() const;

	//! Druh mocnina L2-normy vektoru.
	/*!
	\return Hodnotu \f$\mathbf{||v||^2}=x^2+y^2+z^2\f$.
	*/
	float SqrL2Norm() const;

	//! Normalizace vektoru.
	/*!
	Po proveden operace bude mt vektor jednotkovou dlku.
	*/
	float Normalize();

	//! Vektorov souin.
	/*!
	\param v vektor \f$\mathbf{v}\f$.

	\return Vektor \f$(\mathbf{u}_x \mathbf{v}_z - \mathbf{u}_z \mathbf{v}_y,
	\mathbf{u}_z \mathbf{v}_x - \mathbf{u}_x \mathbf{v}_z,
	\mathbf{u}_x \mathbf{v}_y - \mathbf{u}_y \mathbf{v}_x)\f$.
	*/
	Vector3 CrossProduct( const Vector2 & v ) const;	

	/* projects vector v orhogonally onto the line spanned by this vector */
	Vector2 Project( const Vector2 & v ) const;

	Vector2 Abs() const;

	Vector2 Max( const float a = 0 ) const;

	float max_value() const;

	float min_value() const;

	//! Skalrn souin.
	/*!		
	\return Hodnotu \f$\mathbf{u}_x \mathbf{v}_x + \mathbf{u}_y \mathbf{v}_y + \mathbf{u}_z \mathbf{v}_z)\f$.
	*/
	float DotProduct( const Vector2 & v ) const;

	//! Rotace.
	/*!		
	\return Vektor orotovan kolem potku o \f$\phi\f$ radin v kladnm smyslu.
	*/
	//Vector2 Rotate( const float phi );

	//! Index nejvt sloky vektoru.
	/*!
	\param absolute_value index bude uren podle absolutn hodnoty sloky

	\return Index nejvt sloky vektoru.
	*/
	char LargestComponent( const bool absolute_value = false );	

	void Print() const;

	// --- opertory ------

	friend Vector2 operator-( const Vector2 & v );

	friend Vector2 operator+( const Vector2 & u, const Vector2 & v );
	friend Vector2 operator-( const Vector2 & u, const Vector2 & v );

	friend Vector2 operator*( const Vector2 & v, const float a );	
	friend Vector2 operator*( const float a, const Vector2 & v );
	friend Vector2 operator*( const Vector2 & u, const Vector2 & v );

	friend Vector2 operator/( const Vector2 & v, const float a );

	friend void operator+=( Vector2 & u, const Vector2 & v );
	friend void operator-=( Vector2 & u, const Vector2 & v );
	friend void operator*=( Vector2 & v, const float a );
	friend void operator/=( Vector2 & v, const float a );		
	friend bool operator==( const Vector2 & u, const Vector2 & v );
};

static_assert( sizeof( Vector2 ) == 2 * sizeof( float ), "Vector3 size is not correct" );

#endif // VECTOR2_H_
